import configparser
import os
import ast


class Config(object):
    """
        The settings can be changed by setting up a config file. For an example of a config file, see
         config.cfg in the main-directory. 
    """

    def __init__(self):
        """
            Sets the default values for the project
        """
        self.version = '1.0'
        self.link = 'dashboard'
        self.database_name = 'sqlite:///flask_monitoringdashboard.db'
        self.test_dir = None
        self.username = 'admin'
        self.password = 'admin'
        self.guest_username = 'guest'
        self.guest_password = ['guest_password']
        self.outlier_detection_constant = 2.5
        self.colors = {}
        self.security_token = 'cc83733cb0af8b884ff6577086b87909'
        self.outliers_enabled = True

        # define a custom function to retrieve the session_id or username
        self.get_group_by = None

    def init_from(self, file=None, envvar=None):
        """
            The config_file must at least contains the following variables in section 'dashboard':
            APP_VERSION: the version of the app that you use. Updating the version helps in 
                showing differences in execution times of a function over a period of time.
            CUSTOM_LINK: The dashboard can be visited at localhost:5000/{{CUSTOM_LINK}}.
            DATABASE: Suppose you have multiple projects where you're working on and want to 
                separate the results. Then you can specify different database_names, such that the 
                result of each project is stored in its own database.
            
            Since updating the version in the config-file when updating code isn't very useful, it
            is a better idea to provide the location of the git-folder. From the git-folder. The 
            version automatically retrieved by reading the commit-id (hashed value):
            GIT = If you're using git, then it is easier to set the location to the .git-folder, 
                The location is relative to the config-file.

            USERNAME: for logging into the dashboard, a username and password is required. The
                username can be set using this variable.
            PASSWORD: same as for the username, but this is the password variable.
            GUEST_USERNAME: A guest can only see the results, but cannot configure/download data.
            GUEST_PASSWORD: A guest can only see the results, but cannot configure/download data.

            OUTLIER_DETECTION_CONSTANT: When the execution time is more than this constant *
                average, extra information is logged into the database. A default value for this
                variable is 2.5, but can be changed in the config-file.

            SECURITY_TOKEN: Used for getting the data in /get_json_data/<security_token>

            OUTLIERS_ENABLED: Whether you want the Dashboard to collect extra information about outliers.

            :param file: a string pointing to the location of the config-file
            :param envvar: a string specifying which environment variable holds the config file location
        """

        if envvar:
            file = os.getenv(envvar)
        if not file:
            print("No configuration file specified. Please do so.")
            return

        # When collecting unit test performance results, create log file
        log_dir = os.getenv('DASHBOARD_LOG_DIR')
        if log_dir:
            log = open(log_dir + "endpoint_hits.log", "w")
            log.write("\"time\",\"endpoint\"\n")
            log.close()

        parser = configparser.RawConfigParser()
        try:
            parser.read(file)
            if parser.has_option('dashboard', 'APP_VERSION'):
                self.version = parser.get('dashboard', 'APP_VERSION')
            if parser.has_option('dashboard', 'CUSTOM_LINK'):
                self.link = parser.get('dashboard', 'CUSTOM_LINK')
            if parser.has_option('dashboard', 'DATABASE'):
                self.database_name = parser.get('dashboard', 'DATABASE')
            if parser.has_option('dashboard', 'TEST_DIR'):
                self.test_dir = parser.get('dashboard', 'TEST_DIR')
            if parser.has_option('dashboard', 'OUTLIERS_ENABLED'):
                self.outliers_enabled = parser.get('dashboard', 'OUTLIERS_ENABLED') == "True"

            # For manually defining colors of specific endpoints
            if parser.has_option('dashboard', 'COLORS'):
                self.colors = ast.literal_eval(parser.get('dashboard', 'COLORS'))

            # When the option git is selected, it overrides the given version
            if parser.has_option('dashboard', 'GIT'):
                git = parser.get('dashboard', 'GIT')
                try:
                    # current hash can be found in the link in HEAD-file in git-folder
                    # The file is specified by: 'ref: <location>'
                    git_file = (open(os.path.join(git, 'HEAD')).read().rsplit(': ', 1)[1]).rstrip()
                    # read the git-version
                    self.version = open(git + '/' + git_file).read()
                    # cut version to at most 6 chars
                    self.version = self.version[:6]
                except IOError:
                    print("Error reading one of the files to retrieve the current git-version.")
                    raise

            # provide username and/or password ..
            # .. for admin
            if parser.has_option('dashboard', 'USERNAME'):
                self.username = parser.get('dashboard', 'USERNAME')
            if parser.has_option('dashboard', 'PASSWORD'):
                self.password = parser.get('dashboard', 'PASSWORD')
            # .. for guest (a guest can only see the results, but cannot configure or download any data)
            if parser.has_option('dashboard', 'GUEST_USERNAME'):
                self.guest_username = parser.get('dashboard', 'GUEST_USERNAME')
            if parser.has_option('dashboard', 'GUEST_PASSWORD'):
                self.guest_password = ast.literal_eval(parser.get('dashboard', 'GUEST_PASSWORD'))

            # when an outlier detection constant has been set up:
            if parser.has_option('dashboard', 'OUTLIER_DETECTION_CONSTANT'):
                self.outlier_detection_constant = ast.literal_eval(
                    parser.get('dashboard', 'OUTLIER_DETECTION_CONSTANT'))

            # when a security token is provided:
            if parser.has_option('dashboard', 'security_token'):
                self.security_token = parser.get('dashboard', 'SECURITY_TOKEN')

        except configparser.Error:
            raise



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, login, \
    NAME


class TestResult(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_result_heatmap(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/result/{}/heatmap'.format(NAME)).status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/result/{}/heatmap'.format(NAME)).status_code)

    def test_result_time_per_hour(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/result/{}/time_per_hour'.format(NAME)).status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/result/{}/time_per_hour'.format(NAME)).status_code)

    def test_result_hits_per_hour(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/result/{}/hits_per_hour'.format(NAME)).status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/result/{}/hits_per_hour'.format(NAME)).status_code)

    def test_result_time_per_version_per_user(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/result/{}/time_per_version_per_user'.format(NAME)).status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/result/{}/time_per_version_per_user'.format(NAME)).status_code)

    def test_result_time_per_version_per_ip(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/result/{}/time_per_version_per_ip'.format(NAME)).status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/result/{}/time_per_version_per_ip'.format(NAME)).status_code)

    def test_result_time_per_version(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/result/{}/time_per_version'.format(NAME)).status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/result/{}/time_per_version'.format(NAME)).status_code)

    def test_result_time_per_user(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/result/{}/time_per_user'.format(NAME)).status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/result/{}/time_per_user'.format(NAME)).status_code)

    def test_result_outliers(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        self.assertEqual(302, self.app.get('dashboard/result/{}/outliers'.format(NAME)).status_code)
        login(self.app)
        self.assertEqual(200, self.app.get('dashboard/result/{}/outliers'.format(NAME)).status_code)


import unittest

from flask_monitoringdashboard.test.utils import get_test_app


class TestConfig(unittest.TestCase):

    def test_init(self):
        """
            Test whether the init_form doesn't break when nothing is specified
        """
        import flask_monitoringdashboard as dashboard
        dashboard.config.init_from()



"""
    This file contains all unit tests for the endpoint-table in the database. (Corresponding to the file:
    'dashboard/database/function_calls.py')
    See __init__.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, mean, \
    EXECUTION_TIMES, TIMES, NAME, GROUP_BY, IP


class TestFunctionCall(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_get_reqs_endpoint_day(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import get_reqs_endpoint_day
        result = get_reqs_endpoint_day()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].cnt, len(EXECUTION_TIMES))
        self.assertEqual(result[0].endpoint, NAME)

    def test_add_function_call(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import add_function_call, get_data_per_endpoint
        name2 = 'main2'
        execution_time = 1234
        self.assertNotEqual(NAME, name2, 'Both cannot be equal, otherwise the test will fail')
        self.assertEqual(len(get_data_per_endpoint(name2)), 0)
        add_function_call(execution_time, name2, ip=IP)

        result2 = get_data_per_endpoint(name2)
        self.assertEqual(len(result2), 1)
        self.assertEqual(result2[0].endpoint, name2)
        self.assertEqual(result2[0].execution_time, execution_time)

    def test_get_times(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import get_times
        result = get_times()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].endpoint, NAME)
        self.assertEqual(result[0].count, len(EXECUTION_TIMES))
        self.assertEqual(result[0].average, mean(EXECUTION_TIMES))

    def test_get_data_from(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import get_data_between
        result = get_data_between(TIMES[len(TIMES)-1], TIMES[len(TIMES)-1])
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].endpoint, NAME)
        self.assertEqual(result[0].execution_time, EXECUTION_TIMES[len(TIMES)-1])
        self.assertEqual(result[0].time, TIMES[len(TIMES)-1])
        self.assertEqual(result[0].group_by, GROUP_BY)
        self.assertEqual(result[0].ip, IP)

    def test_get_data(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import get_data, config
        result = get_data()
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        for i in range(len(EXECUTION_TIMES)):
            self.assertEqual(result[i].endpoint, NAME)
            self.assertEqual(result[i].execution_time, EXECUTION_TIMES[i])
            self.assertEqual(result[i].time, TIMES[i])
            self.assertEqual(result[i].group_by, GROUP_BY)
            self.assertEqual(result[i].version, config.version)
            self.assertEqual(result[i].ip, IP)

    def test_get_data_per_version(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import get_data_per_version, config
        new_version = 1.1
        self.assertNotEqual(config.version, new_version)
        self.assertEqual(get_data_per_version(new_version), [])

        result = get_data_per_version(config.version)
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        for i in range(len(EXECUTION_TIMES)):
            self.assertEqual(result[i].execution_time, EXECUTION_TIMES[i])
            self.assertEqual(result[i].version, config.version)

    def test_get_versions(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import get_versions, config
        result = get_versions()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].version, config.version)
        self.assertEqual(result[0].startedUsingOn, TIMES[0])

    def test_get_data_per_endpoint(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import get_data_per_endpoint
        new_name = 'main2'
        self.assertNotEqual(NAME, new_name)
        self.assertEqual(get_data_per_endpoint(new_name), [])

        result = get_data_per_endpoint(NAME)
        self.assertEqual(len(result), len(EXECUTION_TIMES))
        for i in range(len(EXECUTION_TIMES)):
            self.assertEqual(result[i].endpoint, NAME)
            self.assertEqual(result[i].execution_time, EXECUTION_TIMES[i])

    def test_get_endpoints(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.function_calls import get_endpoints
        result = get_endpoints()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].endpoint, NAME)
        self.assertEqual(result[0].cnt, len(EXECUTION_TIMES))


"""
Contains all functions that returns results of all monitor_rules
"""
from flask_monitoringdashboard.database import session_scope, MonitorRule


def get_monitor_rules():
    """ Return all monitor rules that are currently being monitored"""
    with session_scope() as db_session:
        result = db_session.query(MonitorRule).filter(MonitorRule.monitor).all()
        db_session.expunge_all()
        return result


def get_monitor_names():
    """ Return all names of monitor rules that are currently being monitored"""
    with session_scope() as db_session:
        result = db_session.query(MonitorRule.endpoint).filter(MonitorRule.monitor).all()
        db_session.expunge_all()
        return result


def reset_monitor_endpoints():
    """ Update all monitor rules in the database and set them to false. """
    with session_scope() as db_session:
        db_session.query(MonitorRule).update({MonitorRule.monitor: False})


def get_monitor_data():
    """
    Returns all data in the rules-table. This table contains which endpoints are being
    monitored and which are not.
    :return: all data from the database in the rules-table.
    """
    with session_scope() as db_session:
        result = db_session.query(MonitorRule.endpoint,
                                  MonitorRule.last_accessed,
                                  MonitorRule.monitor,
                                  MonitorRule.time_added,
                                  MonitorRule.version_added).all()
        db_session.expunge_all()
        return result



import datetime
import math
import plotly
import plotly.graph_objs as go

from flask import session, url_for, render_template, request
from flask_wtf import FlaskForm
from werkzeug.routing import BuildError
from wtforms import SelectMultipleField, SubmitField

from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.database import FunctionCall, Outlier
from flask_monitoringdashboard.database.endpoint import get_endpoint_column, get_endpoint_results, get_monitor_rule, \
    get_line_results, get_all_measurement_per_column, get_endpoint_column_user_sorted, get_all_measurement
from flask_monitoringdashboard.database.function_calls import get_versions
from flask_monitoringdashboard.security import secure
from flask_monitoringdashboard.routings.measurements import get_heatmap
from flask_monitoringdashboard.database.outlier import get_outliers_sorted
from flask_monitoringdashboard.colors import get_color

# Constants
BUBBLE_SIZE_RATIO = 1250


@blueprint.route('/result/<end>/heatmap')
@secure
def result_heatmap(end):
    rule = get_monitor_rule(end)
    url = get_url(end)
    return render_template('endpoint/plotly.html', link=config.link, session=session, rule=rule, url=url,
                           graph=get_heatmap(end), end=end, index=0)


@blueprint.route('/result/<end>/time_per_hour')
@secure
def result_time_per_hour(end):
    rule = get_monitor_rule(end)
    url = get_url(end)
    return render_template('endpoint/plotly.html', link=config.link, session=session, rule=rule, url=url,
                           graph=get_time_per_hour(end), end=end, index=1)


@blueprint.route('/result/<end>/hits_per_hour')
@secure
def result_hits_per_hour(end):
    rule = get_monitor_rule(end)
    url = get_url(end)
    return render_template('endpoint/plotly.html', link=config.link, session=session, rule=rule, url=url,
                           graph=get_hits_per_hour(end), end=end, index=2)


@blueprint.route('/result/<end>/time_per_version_per_user', methods=['GET', 'POST'])
@secure
def result_time_per_version_per_user(end):
    rule = get_monitor_rule(end)
    url = get_url(end)
    graph, form = get_time_per_version_per_user(end, get_versions(end))
    return render_template('endpoint/time_per_user.html', link=config.link, session=session, rule=rule, url=url,
                           graph=graph, form=form, end=end, index=3)


@blueprint.route('/result/<end>/time_per_version_per_ip', methods=['GET', 'POST'])
@secure
def result_time_per_version_per_ip(end):
    rule = get_monitor_rule(end)
    url = get_url(end)
    graph, form = get_time_per_version_per_ip(end, get_versions(end))
    return render_template('endpoint/time_per_user.html', link=config.link, session=session, rule=rule, url=url,
                           graph=graph, form=form, end=end, index=4)


@blueprint.route('/result/<end>/time_per_version')
@secure
def result_time_per_version(end):
    rule = get_monitor_rule(end)
    url = get_url(end)
    return render_template('endpoint/plotly.html', link=config.link, session=session, rule=rule, url=url,
                           graph=get_time_per_version(end, get_versions(end)), end=end, index=5)


@blueprint.route('/result/<end>/time_per_user', methods=['GET', 'POST'])
@secure
def result_time_per_user(end):
    rule = get_monitor_rule(end)
    url = get_url(end)
    graph, form = get_time_per_user(end)
    return render_template('endpoint/time_per_user.html', link=config.link, session=session, rule=rule, url=url,
                           graph=graph, form=form, end=end, index=6)


@blueprint.route('/result/<end>/outliers')
@secure
def result_outliers(end):
    rule = get_monitor_rule(end)
    url = get_url(end)
    return render_template('endpoint/outliers.html', link=config.link, session=session, rule=rule, url=url,
                           end=end, index=7, table=get_outliers_sorted(end, Outlier.execution_time))


def formatter(ms):
    """
    formats the ms into seconds and ms
    :param ms: the number of ms
    :return: a string representing the same amount, but now represented in seconds and ms.
    """
    sec = int(ms) // 1000
    ms = int(ms) % 1000
    if sec == 0:
        return '{0}ms'.format(ms)
    return '{0}.{1}s'.format(sec, ms)


def get_url(end):
    """
    Returns the URL if possible.
    URL's that require additional arguments, like /static/<file> cannot be retrieved.
    :param end: the endpoint for the url.
    :return: the url_for(end) or None,
    """
    try:
        return url_for(end)
    except BuildError:
        return None


def get_time_per_hour(end):
    data = get_line_results(end)

    # Find the last date which contains data
    max_time = max([d.newTime for d in data])
    max_date = datetime.datetime.strptime(max_time, '%Y-%m-%d %H:%M:%S')
    max_date += datetime.timedelta(minutes=30)

    trace1 = go.Bar(
        x=[d.newTime for d in data],
        y=[d.min for d in data],
        name='Minimum'
    )

    trace2 = go.Bar(
        x=[d.newTime for d in data],
        y=[d.avg for d in data],
        name='Average'
    )

    trace3 = go.Bar(
        x=[d.newTime for d in data],
        y=[d.max for d in data],
        name='Maximum'
    )

    graph = [trace3, trace2, trace1]

    layout = go.Layout(
        autosize=True,
        height=600,
        title='Execution time (minimum, average and maximum) per hour',
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=True,
        barmode='overlay',
        xaxis=go.XAxis(
            title='Date',
            range=[max_date - datetime.timedelta(days=2), max_date]
        ),
        yaxis=go.YAxis(
            title='Execution time (ms)'
        )
    )
    return plotly.offline.plot(go.Figure(data=graph, layout=layout), output_type='div', show_link=False)


def get_hits_per_hour(end):
    data = get_line_results(end)

    # Find the last date which contains data
    max_time = max([d.newTime for d in data])
    max_date = datetime.datetime.strptime(max_time, '%Y-%m-%d %H:%M:%S')
    max_date += datetime.timedelta(minutes=30)

    graph = [go.Bar(
        x=[d.newTime for d in data],
        y=[d.count for d in data],
        marker=dict(color=get_color(end))
    )]

    layout = go.Layout(
        autosize=True,
        height=600,
        title='Number of hits per hour',
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        xaxis=go.XAxis(
            title='Date',
            range=[max_date - datetime.timedelta(days=2), max_date]
        ),
        yaxis=go.YAxis(title='Hits')
    )
    return plotly.offline.plot(go.Figure(data=graph, layout=layout), output_type='div', show_link=False)


def get_time_per_version_per_user(end, versions):
    user_data = {}
    data = [t.execution_time for t in get_all_measurement(end)]
    # compute the average for determining the size of the bubbles in the plot
    average = math.sqrt(sum(data) / len(data)) / BUBBLE_SIZE_RATIO

    for d in [str(c.group_by) for c in get_endpoint_column(end, FunctionCall.group_by)]:
        user_data[d] = {}
        for v in versions:
            user_data[d][v.version] = 0

    form, selection = get_form(user_data)
    # fill the rows with data
    for d in get_endpoint_results(end, FunctionCall.group_by):
        user_data[str(d.group_by)][d.version] = d.average

    db_data = [str(c.group_by) for c in get_endpoint_column(end, FunctionCall.group_by)]
    trace = []
    for d in db_data:  # iterate through all (unique) ip addresses
        if selection == [] or d in selection:
            data = []
            for version in versions:
                data.append(user_data[d][version.version])

            hover_text = []
            for i in range(len(data)):
                hover_text.append('Version: ' + versions[i].version + '<br>Time: ' + formatter(data[i]))

            trace.append(go.Scatter(
                x=["{0}<br>{1}".format(v.version, v.startedUsingOn.strftime("%b %d %H:%M")) for v in versions],
                hovertext=hover_text,
                y=[d] * len(versions),
                name=d,
                mode='markers',
                marker=dict(
                    color=[get_color(d)] * len(versions),
                    size=[math.sqrt(d) for d in data],
                    sizeref=average,
                    sizemode='area'
                )
            ))

    layout = go.Layout(
        autosize=True,
        height=350 + 40 * len(trace),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Average execution time for every user per version',
        xaxis=dict(
            title='Versions',
            type='category'
        ),
        yaxis=dict(
            type='category',
            title='Users',
            autorange='reversed'
        ),
        margin=go.Margin(
            l=200,
            b=200
        )
    )

    return plotly.offline.plot(go.Figure(data=trace, layout=layout), output_type='div', show_link=False), form


def get_form(data):
    # create a form to make a selection
    choices = []
    for d in list(data):
        choices.append((d, d))

    class SelectionForm(FlaskForm):
        selection = SelectMultipleField(
            'Pick Things!',
            choices=choices,
        )
        submit = SubmitField('Render graph')

    form = SelectionForm(request.form)
    selection = []
    if request.method == 'POST':
        selection = [str(item) for item in form.data['selection']]

    return form, selection


def get_time_per_version_per_ip(end, versions):
    ip_data = {}
    data = [t.execution_time for t in get_all_measurement(end)]
    # compute the average for determining the default size
    average = math.sqrt(sum(data) / len(data)) / 1250
    for d in [str(c.ip) for c in get_endpoint_column(end, FunctionCall.ip)]:
        ip_data[d] = {}
        for v in versions:
            ip_data[d][v.version] = 0

    form, selection = get_form(ip_data)

    # fill the rows with data
    for d in get_endpoint_results(end, FunctionCall.ip):
        ip_data[str(d.ip)][d.version] = d.average

    db_data = [str(c.ip) for c in get_endpoint_column(end, FunctionCall.ip)]
    trace = []
    for d in db_data:  # iterate through all (unique) ip addresses
        if selection == [] or d in selection:
            data = []
            for version in versions:
                data.append(ip_data[d][version.version])

            hover_text = []
            for i in range(len(data)):
                hover_text.append('Version: ' + versions[i].version + '<br>Time: ' + formatter(data[i]))

            trace.append(go.Scatter(
                x=["{0}<br>{1}".format(v.version, v.startedUsingOn.strftime("%b %d %H:%M")) for v in versions],
                hovertext=hover_text,
                y=[d] * len(versions),
                name=d,
                mode='markers',
                marker=dict(
                    color=[get_color(d)] * len(versions),
                    size=[math.sqrt(d) for d in data],
                    sizeref=average,
                    sizemode='area'
                )
            ))

    layout = go.Layout(
        autosize=True,
        height=350 + 40 * len(trace),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Average execution time for every IP-address per version',
        xaxis=dict(
            title='Versions',
            type='category'
        ),
        yaxis=dict(
            type='category',
            title='IP-addresses',
            autorange='reversed'
        ),
        margin=go.Margin(
            l=200,
            b=200
        )
    )

    return plotly.offline.plot(go.Figure(data=trace, layout=layout), output_type='div', show_link=False), form


def get_time_per_version(end, versions):
    data = []
    for v in versions:
        values = [str(c.execution_time) for c in
                  get_all_measurement_per_column(endpoint=end, column=FunctionCall.version, value=v.version)]

        data.append(go.Box(
            x=values,
            marker=dict(
                color=get_color(end)
            ),
            name="{0} {1}".format(v.version, v.startedUsingOn.strftime("%b %d %H:%M"))))

    layout = go.Layout(
        autosize=True,
        height=350 + 40 * len(versions),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Execution time for every version',
        xaxis=dict(title='Execution time (ms)'),
        yaxis=dict(
            title='Version',
            autorange='reversed'
        ),
        margin=go.Margin(
            l=200
        )
    )
    return plotly.offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)


def get_time_per_user(end):
    users = [str(c.group_by) for c in get_endpoint_column_user_sorted(endpoint=end, column=FunctionCall.group_by)]
    form, selection = get_form(users)
    data = []
    for user in users:
        if selection == [] or user in selection:
            values = [str(c.execution_time) for c in
                      get_all_measurement_per_column(endpoint=end, column=FunctionCall.group_by, value=user)]
            data.append(go.Box(
                x=values,
                marker=dict(
                    color=get_color(user)
                ),
                name='{0}  -'.format(user)))
    data.reverse()
    layout = go.Layout(
        autosize=True,
        height=350 + 40 * len(data),
        plot_bgcolor='rgba(249,249,249,1)',
        showlegend=False,
        title='Execution time for every user',
        xaxis=dict(title='Execution time (ms)'),
        yaxis=dict(title='User')
    )
    graph = plotly.offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)
    return graph, form



import datetime
import jwt
import pkg_resources
import time

from flask import make_response, render_template, session, request, json, jsonify

from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.database.function_calls import get_data, get_data_between, get_date_of_first_request
from flask_monitoringdashboard.database.monitor_rules import get_monitor_data
from flask_monitoringdashboard.database.tests import add_or_update_test, add_test_result, get_suite_nr
from flask_monitoringdashboard.database.tests_grouped import reset_tests_grouped, add_tests_grouped
from flask_monitoringdashboard.security import admin_secure
# from setup import VERSION

CSV_COLUMNS = ['endpoint', 'execution_time', 'time', 'version', 'group_by', 'ip']


@blueprint.route('/download-csv')
@admin_secure
def download_csv():
    csv = ','.join(CSV_COLUMNS) + '\n'
    for entry in get_data():
        csv += ','.join([str(entry.__getattribute__(c)) for c in CSV_COLUMNS]) + '\n'

    response = make_response(csv)
    response.headers["Content-Disposition"] = "attachment; filename=measurements_{0}.csv".format(
        str(datetime.datetime.now()).replace(" ", "_").replace(":", "-")[:19])
    return response


@blueprint.route('/export-data')
@admin_secure
def export_data():
    csv = [','.join(CSV_COLUMNS)]
    data = get_data()
    for entry in data:
        csv.append(','.join([str(entry.__getattribute__(c)) for c in CSV_COLUMNS]) + '\n')
    return render_template('dashboard/export-data.html', link=config.link, session=session, data=csv)


@blueprint.route('/submit-test-results', methods=['POST'])
def submit_test_results():
    content = request.get_json()['test_runs']
    suite = get_suite_nr()
    for result in content:
        time = datetime.datetime.strptime(result['time'], '%Y-%m-%d %H:%M:%S.%f')
        add_or_update_test(result['name'], time, result['successful'])
        add_test_result(result['name'], result['exec_time'], time, config.version, suite, result['iter'])

    groups = request.get_json()['grouped_tests']
    if groups:
        reset_tests_grouped()
        add_tests_grouped(groups)

    return '', 204


@blueprint.route('/get_json_data', defaults={'time_from': 0})
@blueprint.route('/get_json_data/<time_from>')
@blueprint.route('/get_json_data/<time_from>/<time_to>')
def get_json_data_from(time_from, time_to=None):
    """
    The returned data is the data that is encrypted using a security token. This security token is set in the
    configuration.
    :param time_from: (optional) if specified, only the data-values after this date are returned.
                      input must be an timestamp value (utc) (= integer)
                      :param time_to: (optional) if specified, only the data-values before this date are returned.
                      input must be an timestamp value (utc) (= integer)
    :return: all entries from the database. (endpoint-table)
    """
    data = []
    try:
        time1 = datetime.datetime.utcfromtimestamp(int(time_from))
        time2 = None
        if time_to:
            time2 = datetime.datetime.utcfromtimestamp(int(time_to))
        for entry in get_data_between(time1, time2):
            # nice conversion to json-object
            data.append({
                'endpoint': entry.endpoint,
                'execution_time': entry.execution_time,
                'time': str(entry.time),
                'version': entry.version,
                'group_by': entry.group_by,
                'ip': entry.ip
            })
        return jwt.encode({'data': json.dumps(data)}, config.security_token, algorithm='HS256')
    except ValueError as e:
        return 'ValueError: {}'.format(e)


@blueprint.route('/get_json_monitor_rules')
def get_json_monitor_rules():
    """
    The returned data is the data that is encrypted using a security token. This security token is set in the
    configuration.
    :return: all entries from the database (rules-table)
    """
    data = []
    try:
        for entry in get_monitor_data():
            # nice conversion to json-object
            data.append({
                'endpoint': entry.endpoint,
                'last_accessed': str(entry.last_accessed),
                'monitor': entry.monitor,
                'time_added': str(entry.time_added),
                'version_added': entry.version_added
            })
        return jwt.encode({'data': json.dumps(data)}, config.security_token, algorithm='HS256')
    except ValueError as e:
        return 'ValueError: {}'.format(e)


@blueprint.route('/get_json_details')
def get_json_details():
    """
    Some details about the deployment, such as the current version, etc...
    :return: a json-object with the details.
    """
    version = pkg_resources.require("Flask-MonitoringDashboard")[0].version
    return jsonify({'version': version,
                    'first_request': get_date_of_first_request()})



"""
Contains all functions that access a single endpoint
"""
from sqlalchemy import func, text, asc
from sqlalchemy.orm.exc import NoResultFound
import datetime

from flask_monitoringdashboard import config
from flask_monitoringdashboard.database import session_scope, FunctionCall, MonitorRule


def get_line_results(endpoint):
    """
    Returns simple statistics, such as average, minimum, maximum and count from a given endpoint
    :param endpoint: the name of the endpoint
    :return: simple statistics, as described above
    """
    with session_scope() as db_session:
        query = text("""select
                datetime(CAST(strftime('%s', time)/3600 AS INT)*3600, 'unixepoch') as newTime, 
                avg(execution_time) as avg,
                min(execution_time) as min,
                max(execution_time) as max,
                count(execution_time) as count
            from functioncalls 
            where endpoint=:val group by newTime""")
        result = db_session.execute(query, {'val': endpoint})
        data = result.fetchall()
        return data


def get_num_requests(endpoint):
    """ Returns a list with all dates on which an endpoint is accessed.
        :param endpoint: if None, the result is the sum of all endpoints
    """
    with session_scope() as db_session:
        query = text("""select
                datetime(CAST(strftime('%s', time)/3600 AS INT)*3600, 'unixepoch') AS newTime,
                count(execution_time) as count
                from functioncalls
                where (endpoint=:val OR :val='None') group by newTime""")
        result = db_session.execute(query, {'val': str(endpoint)})
        data = result.fetchall()
        return data


def get_endpoint_column(endpoint, column):
    """ Returns a list of entries from column in which the endpoint is involved. """
    with session_scope() as db_session:
        result = db_session.query(column,
                                  func.min(FunctionCall.time).label('startedUsingOn')). \
            filter(FunctionCall.endpoint == endpoint). \
            group_by(column).order_by(asc('startedUsingOn')).all()
        db_session.expunge_all()
        return result


def get_endpoint_column_user_sorted(endpoint, column):
    """ Returns a list of entries from column in which the endpoint is involved. """
    with session_scope() as db_session:
        result = db_session.query(column). \
            filter(FunctionCall.endpoint == endpoint). \
            group_by(column).order_by(asc(column)).all()
        db_session.expunge_all()
        return result


def get_endpoint_results(endpoint, column):
    """ Returns a list of entries with measurements in which the endpoint is involved.
    The entries are grouped by their version and the given column. """
    with session_scope() as db_session:
        result = db_session.query(FunctionCall.version,
                                  column,
                                  func.count(FunctionCall.execution_time).label('count'),
                                  func.avg(FunctionCall.execution_time).label('average')
                                  ).filter(FunctionCall.endpoint == endpoint). \
            group_by(FunctionCall.version, column).all()
        db_session.expunge_all()
        return result


def get_monitor_rule(endpoint):
    """ Get the MonitorRule from a given endpoint. If no value is found, a new value 
    is added to the database and the function is called again (recursively). """
    try:
        with session_scope() as db_session:
            result = db_session.query(MonitorRule). \
                filter(MonitorRule.endpoint == endpoint).one()
            # for using the result when the session is closed, use expunge
            db_session.expunge(result)
            return result
    except NoResultFound:
        with session_scope() as db_session:
            db_session.add(
                MonitorRule(endpoint=endpoint, version_added=config.version, time_added=datetime.datetime.now()))

        # return new added row
        return get_monitor_rule(endpoint)


def update_monitor_rule(endpoint, value):
    """ Update the value of a specific monitor rule. """
    with session_scope() as db_session:
        db_session.query(MonitorRule).filter(MonitorRule.endpoint == endpoint). \
            update({MonitorRule.monitor: value})


def get_all_measurement(endpoint):
    """Return all entries with measurements from a given endpoint. Used for creating a histogram. """
    with session_scope() as db_session:
        result = db_session.query(FunctionCall).filter(FunctionCall.endpoint == endpoint).all()
        db_session.expunge_all()
        return result


def get_all_measurement_per_column(endpoint, column, value):
    """Return all entries with measurements from a given endpoint for which the column has a specific value.
    Used for creating a box plot. """
    with session_scope() as db_session:
        result = db_session.query(FunctionCall).filter(FunctionCall.endpoint == endpoint, column == value).all()
        db_session.expunge_all()
        return result


def get_last_accessed_times():
    """ Returns a list of all endpoints and their last accessed time. """
    with session_scope() as db_session:
        result = db_session.query(MonitorRule.endpoint, MonitorRule.last_accessed).all()
        db_session.expunge_all()
        return result


def update_last_accessed(endpoint, value):
    """ Updates the timestamp of last access of the endpoint. """
    with session_scope() as db_session:
        db_session.query(MonitorRule).filter(MonitorRule.endpoint == endpoint). \
            update({MonitorRule.last_accessed: value})
